import json
from typing import List

from aiohttp import ClientSession

from pyamplipi.client import Client
from pyamplipi.models import Group, Stream, SourceUpdate, MultiZoneUpdate, ZoneUpdate, \
    GroupUpdate, StreamUpdate, Announcement, Status, Source, Zone, Preset


class AmpliPi:
    def __init__(
            self,
            endpoint: str,
            timeout: 10,
            http_session: ClientSession = None,
            verify_ssl: bool = False,
            disable_insecure_warning: bool = True,
    ):
        self._client = Client(
            endpoint,
            timeout,
            http_session,
            verify_ssl,
            disable_insecure_warning,
        )

    async def get_status(self) -> Status:
        response = await self._client.get('')
        return Status.parse_obj(response)

    async def get_sources(self) -> List[Source]:
        response = await self._client.get('sources')
        return [Source.parse_obj(source) for source in response['sources']]

    async def get_source(self, source_id: int) -> Source:
        response = await self._client.get(f'sources/{source_id}')
        return Source.parse_obj(response)

    async def set_source(self, source_id: int, source_update: SourceUpdate) -> Status:
        response = await self._client.patch(f'sources/{source_id}', source_update.json())
        return Status.parse_obj(response)

    async def get_zone(self, zone_id: int) -> Zone:
        response = await self._client.get(f'zones/{zone_id}')
        return Zone.parse_obj(response)

    async def get_zones(self) -> List[Zone]:
        response = await self._client.get('zones')
        return [Zone.parse_obj(zone) for zone in response['zones']]

    async def set_zones(self, zone_update: MultiZoneUpdate) -> Status:
        response = await self._client.patch('zones', zone_update.json())
        return Status.parse_obj(response)

    async def set_zone(self, zone_id: int, zone_update: ZoneUpdate) -> Status:
        response = await self._client.patch(f'zones/{zone_id}',
                                            zone_update.json())
        return Status.parse_obj(response)

    async def create_group(self, new_group: Group) -> Group:
        response = await self._client.post('group', new_group.json())
        return Group.parse_obj(response)

    async def get_groups(self) -> List[Group]:
        response = await self._client.get('groups')
        return [Group.parse_obj(group) for group in response['groups']]

    async def get_group(self, group_id) -> Group:
        response = await self._client.get(f'groups/{group_id}')
        return Group.parse_obj(response)

    async def delete_group(self, group_id):
        response = await self._client.delete(f'groups/{group_id}')
        return Status.parse_obj(response)

    async def set_group(self, group_id, update: GroupUpdate) -> Status:
        response = await self._client.patch(f'groups/{group_id}', update.json())
        return Status.parse_obj(response)

    async def get_streams(self) -> List[Stream]:
        response = await self._client.get('streams')
        return [Stream.parse_obj(stream) for stream in response['streams']]

    async def get_stream(self, stream_id: int) -> Stream:
        response = await self._client.get(f'streams/{stream_id}')
        return Stream.parse_obj(response)

    async def play_stream(self, stream_id: int) -> Stream:
        response = await self._client.post(f'streams/{stream_id}/play')
        return Status.parse_obj(response)

    async def pause_stream(self, stream_id: int) -> Stream:
        response = await self._client.post(f'streams/{stream_id}/pause')
        return Status.parse_obj(response)

    async def previous_stream(self, stream_id: int) -> Stream:
        response = await self._client.post(f'streams/{stream_id}/prev')
        return Status.parse_obj(response)

    async def next_stream(self, stream_id: int) -> Stream:
        response = await self._client.post(f'streams/{stream_id}/next')
        return Status.parse_obj(response)

    async def stop_stream(self, stream_id: int) -> Stream:
        response = await self._client.post(f'streams/{stream_id}/stop')
        return Stream.parse_obj(response)

    async def create_stream(self, new_stream: Stream) -> Stream:
        response = await self._client.post(f'streams', new_stream.json())
        return Status.parse_obj(response)

    async def delete_stream(self, stream_id: int) -> Status:
        response = await self._client.delete(f'streams/{stream_id}')
        return Status.parse_obj(response)

    async def set_stream(self, stream_id: int, update: StreamUpdate) -> Stream:
        response = await self._client.post(f'streams/{stream_id}', update.json())
        return Stream.parse_obj(response)

    async def announce(self, announcement: Announcement) -> Status:
        response = await self._client.post('announce', announcement.json())
        return Status.parse_obj(response)

    async def get_presets(self) -> List[Preset]:
        response = await self._client.get('presets')
        return [Preset.parse_obj(preset) for preset in response['presets']]

    async def close(self):
        await self._client.close()



from http.client import responses
from json.decoder import JSONDecodeError
from urllib.parse import urljoin

import requests
from aiohttp import ClientSession, ClientResponse
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

from .error import AccessDeniedError, APIError, AmpliPiUnreachableError


def headers_or_default(headers=None):
    if headers is None:
        return {
            'Accept-Content': 'application/json',
            'Content-Type': 'application/json',
            'User-Agent': 'pyamplipi-client'
        }
    return headers


class Client(object):
    def __init__(
            self,
            endpoint: str,
            timeout: int = 10,
            http_session: ClientSession = None,
            verify_ssl: bool = False,
            disable_insecure_warning: bool = True,
    ) -> None:

        if disable_insecure_warning:
            disable_warnings(InsecureRequestWarning)

        self._endpoint = self._parse_endpoint(endpoint)
        self._timeout = timeout
        self._http_session = http_session if http_session else ClientSession()
        self._http_session.verify = verify_ssl

    @staticmethod
    def _parse_endpoint(endpoint: str) -> str:

        if not endpoint.endswith("api") and not endpoint.endswith("/"):
            endpoint += "/api/"
        elif endpoint.endswith("api"):
            endpoint += "/"
        elif endpoint.endswith("/") and not endpoint.endswith("api/"):
            endpoint += "api/"

        return endpoint

    @staticmethod
    async def _handle_error(response: ClientResponse) -> None:
        if response.status == 404:
            raise APIError(
                "The url {} returned error 404".format(response.url)
            )

        if response.status == 401 or response.status == 403:

            try:
                response_json = await response.json()
            except Exception:
                raise AccessDeniedError(response.url)
            else:
                raise AccessDeniedError(
                    response.url,
                    response_json.get("error"),
                    response_json.get("message"),
                )

        body = await response.text()

        if body is not None and len(body) > 0:
            raise APIError(
                "API returned status code '{}: {}' with body: {}".format(
                    response.status,
                    responses.get(response.status),
                    body,
                )
            )
        else:
            raise APIError(
                "API returned status code '{}: {}' ".format(
                    response.status, responses.get(response.status)
                )
            )

    async def _process_response(self, response: ClientResponse) -> dict:

        if response.status >= 400:
            # API returned some sort of error that must be handled
            await self._handle_error(response)

        try:
            response_json = await response.json()
        except JSONDecodeError:
            raise APIError(
                "Error while decoding json of response: {}".format(response.text())
            )

        if response_json is None:
            return {}

        # Newer versions of the powerwall do not return such values anymore
        # Kept for backwards compability or if the API changes again
        if "error" in response_json:
            raise APIError(response_json["error"])

        return response_json

    def url(self, path: str):
        return urljoin(self._endpoint, path)

    async def delete(self, path: str, body=None, headers=None) -> dict:
        try:
            async with self._http_session.delete(
                    url=self.url(path),
                    data=body,
                    timeout=self._timeout,
                    headers=headers_or_default(headers),
            ) as response:
                return await self._process_response(response)
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
        ) as e:
            raise AmpliPiUnreachableError(e)

    async def patch(self, path: str, body=None, headers=None) -> dict:
        try:
            async with self._http_session.patch(
                    url=self.url(path),
                    data=body,
                    timeout=self._timeout,
                    headers=headers_or_default(headers),
            ) as response:
                return await self._process_response(response)
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
        ) as e:
            raise AmpliPiUnreachableError(e)

    async def post(self, path: str, body=None, headers=None) -> dict:
        try:
            async with self._http_session.post(
                    url=self.url(path),
                    data=body,
                    timeout=self._timeout,
                    headers=headers_or_default(headers),
            ) as response:
                return await self._process_response(response)
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
        ) as e:
            raise AmpliPiUnreachableError(e)

    async def get(self, path: str, headers=None) -> dict:
        try:
            async with self._http_session.get(
                    url=self.url(path),
                    timeout=self._timeout,
                    headers=headers_or_default(headers),
            ) as response:
                return await self._process_response(response)
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
        ) as e:
            raise AmpliPiUnreachableError(e)

    async def close(self):
        # Close the HTTP Session
        # THis method is required for testing, so python doesn't complain about unclosed resources
        await self._http_session.close()



"""AmpliPi Data Models - Extracted from the amplipi repo.
"""

from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel


class SourceInfo(BaseModel):
    name: str
    state: str  # paused, playing, stopped, unknown, loading ???
    artist: Optional[str]
    track: Optional[str]
    album: Optional[str]
    station: Optional[str]  # name of radio station
    img_url: Optional[str]
    supported_cmds: List[str] = []


class Source(BaseModel):
    """ An audio source """
    id: Optional[int]
    name: str
    input: str
    info: Optional[SourceInfo]  # Additional info about the current audio playing from the stream (generated during
    # playback')


class SourceUpdate(BaseModel):
    """ Partial reconfiguration of an audio Source """
    name: Optional[str]
    input: Optional[str]  # 'None', 'local', 'stream=ID'

class SourceUpdateWithId(SourceUpdate):
    """ Partial reconfiguration of a specific audio Source """
    id: int


class Zone(BaseModel):
    """ Audio output to a stereo pair of speakers, typically belonging to a room """
    id: Optional[int]
    name: str
    source_id: int
    mute: bool
    vol: int
    vol_f: float
    vol_min: int
    vol_max: int
    disabled: bool


class ZoneUpdate(BaseModel):
    """ Reconfiguration of a Zone """
    name: Optional[str]
    source_id: Optional[int]
    mute: Optional[bool]
    vol: Optional[int]
    vol_f: Optional[float]
    vol_min: Optional[int]
    vol_max: Optional[int]
    disabled: Optional[bool]


class ZoneUpdateWithId(ZoneUpdate):
    """ Reconfiguration of a specific Zone """
    id: int


class MultiZoneUpdate(BaseModel):
    """ Reconfiguration of multiple zones specified by zone_ids and group_ids """
    zones: Optional[List[int]]
    groups: Optional[List[int]]
    update: ZoneUpdate


class Group(BaseModel):
    """ A group of zones that can share the same audio input and be controlled as a group ie. Upstairs. Volume, mute,
    and source_id fields are aggregates of the member zones."""
    id: Optional[int]
    name: str
    source_id: Optional[int]
    zones: List[int]
    mute: Optional[bool]
    vol_delta: Optional[int]
    vol_f: Optional[float]


class GroupUpdate(BaseModel):
    """ Reconfiguration of a Group """
    name: Optional[str]
    source_id: Optional[int]
    zones: Optional[List[int]]
    mute: Optional[bool]
    vol_delta: Optional[int]
    vol_f: Optional[float]


class GroupUpdateWithId(GroupUpdate):
    """ Reconfiguration of a specific Group """
    id: int


class Stream(BaseModel):
    """ Digital stream such as Pandora, AirPlay or Spotify """
    id: Optional[int]
    name: str
    type: str
    user: Optional[str]
    password: Optional[str]
    station: Optional[str]
    url: Optional[str]
    logo: Optional[str]
    freq: Optional[str]
    client_id: Optional[str]
    token: Optional[str]


class StreamUpdate(BaseModel):
    """ Reconfiguration of a Stream """
    name: str
    user: Optional[str]
    password: Optional[str]
    station: Optional[str]
    url: Optional[str]
    logo: Optional[str]
    freq: Optional[str]


class StreamCommand(str, Enum):
    PLAY = 'play'
    PAUSE = 'pause'
    NEXT = 'next'
    PREV = 'prev'
    STOP = 'stop'
    LOVE = 'love'
    BAN = 'ban'
    SHELVE = 'shelve'


class PresetState(BaseModel):
    """ A set of partial configuration changes to make to sources, zones, and groups """
    sources: Optional[List[SourceUpdateWithId]]
    zones: Optional[List[ZoneUpdateWithId]]
    groups: Optional[List[GroupUpdateWithId]]


class Command(BaseModel):
    """ A command to execute on a stream """
    stream_id: int
    cmd: str


class Preset(BaseModel):
    id: Optional[int]
    name: str
    state: Optional[PresetState]
    commands: Optional[List[Command]]
    last_used: Union[int, None] = None


class PresetUpdate(BaseModel):
    name: Optional[str]
    state: Optional[PresetState]
    commands: Optional[List[Command]]


class Announcement(BaseModel):
    media: str
    vol: Optional[int]
    vol_f: float
    source_id: int
    zones: Optional[List[int]]
    groups: Optional[List[int]]


class Info(BaseModel):
    config_file: str = 'Uknown'
    version: str = 'Unknown'
    mock_ctrl: bool = False
    mock_streams: bool = False


class Status(BaseModel):
    sources: List[Source] = []
    zones: List[Zone] = []
    groups: List[Group] = []
    streams: List[Stream] = []
    presets: List[Preset] = []
    info: Optional[Info]


class AppSettings(BaseModel):
    mock_ctrl: bool = True
    mock_streams: bool = True
    config_file: str = 'house.json'
    delay_saves: bool = True



class AmpliPiError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class APIError(AmpliPiError):
    def __init__(self, error):
        super().__init__("AmpliPi api error: {}".format(error))


class MissingAttributeError(APIError):
    def __init__(self, response: dict, attribute: str, url: str = None):
        self.response = response
        self.attribute = attribute
        self.url = url

        if url is None:
            super().__init__(
                "The attribute '{}' is expected in the response but is missing.".format(
                    attribute
                )
            )
        else:
            super().__init__(
                "The attribute '{}' is expected in the response for '{}' but is missing.".format(
                    attribute, url
                )
            )


class AmpliPiUnreachableError(AmpliPiError):
    def __init__(self, reason=None):
        msg = "AmpliPi is unreachable"
        self.reason = reason
        if reason is not None:
            msg = "{}: {}".format(msg, reason)
        super().__init__(msg)


class AccessDeniedError(AmpliPiError):
    def __init__(self, resource, error=None, message=None):
        self.resource = resource
        self.error = error
        self.message = message
        msg = "Access denied for resource {}".format(resource)
        if error is not None:
            if message is not None:
                msg = "{}: {}: {}".format(msg, error, message)
            else:
                msg = "{}: {}".format(msg, error)
        super().__init__(msg)



# This is a sample Python script.

# Press ⌃R to execute it or replace it with your code.
# Press Double ⇧ to search everywhere for classes, files, tool windows, actions, and settings.
import asyncio
import logging

from pyamplipi.amplipi import AmpliPi
from pyamplipi.models import ZoneUpdate, ZoneUpdateWithId, SourceUpdate

_LOGGER = logging.getLogger(__name__)


async def get_status():
    amp = AmpliPi(
        "http://amplipi-dev.local:5000/api",
        10,
    )
    result = await amp.get_status()
    source = result.sources.pop()
    update = await amp.set_source(source.id, SourceUpdate(
        mute=not source.mute,
        vol_delta=.01
    ))
    _LOGGER.info(update)
    await amp.close()


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(get_status())

# See PyCharm help at https://www.jetbrains.com/help/pycharm/



#!/usr/bin/env python

import os
import sys

VERSION = '0.4.8'

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

license = """
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

setup(name='pyamplipi',
      version=VERSION,
      description='Python API for interacting with the AmpliPi Multizone Audio Controller',
      url='https://github.com/brianhealey/pyamplipi',
      download_url='https://github.com/brianhealey/pyamplipi/archive/{}.tar.gz'.format(VERSION),
      author='HeeHee Software',
      author_email='brian.healey@gmail.com',
      license='GPL',
      install_requires=[],
      packages=['pyamplipi'],
      classifiers=['Development Status :: 4 - Beta',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6'],
      zip_safe=True)
